// 自动问答核心代码
// 使用逻辑
// 1. 传入一个剧本数组，一个用户信息对象，创建一个上下文对象
// 2. 同时获取到2个函数
// 1. set_select(select) 传入剧本的选择，获取剧本相关的结束语
// 2. get_next() 获取下一个剧本，返回一个对象，包含问题和选项,如果没有下一个剧本，返回null

/* screenplays 结构
 [{
    condition:{} // 条件对象
    question:"" // 问题
    options:[{ // 选项
        text:"" // 选项文本
        key:"" // 选项key,用于设置上下文,key一定要求全剧本唯一
    }],
    end:[{
        condition:{} // 结束语条件
        text:"" // 结束语文本
    }], // 结束语
    key:"" // 剧本key,用于设置上下文,key一定要求全剧本唯一
},]
*/
/*
profile 结构
{'path_acne': 
{
    'tzone_acne': 0, 
    'cheek_acne': 800, 
    'labiomental_acne': 100
},
 'perioral_acne': True, 
 'gender': 'female',
 'name': '小明',
 'age': 18,
  'feedback': 
  {
    'repeated_acne': False, 
    'menstruate_acne': False, 
    'neck_acne': False, 
    'inplace_acne': False,
    'stop_consult': False
   }, 
   'oil': 100, 
   'sensitive': 600
}
*/
// 传入一个剧本数组,用户信息，返回一个上下文对象，用于后续的问答
function match_condition(context, condition) {
    // 判断条件是否满足
    // context是上下文对象，condition是条件对象
    // 返回true或false
    for (let key in condition) {
        if (key === '$or') {
            let have_true = false;
            for (let c of condition[key]) {
                if (match_condition(context, c)) {
                    have_true = true;
                    break;
                }
            }
            if (have_true) {
                continue
            }
            return false
        } else if (key === '$and') {
            for (let c of condition[key]) {
                if (!match_condition(context, c)) {
                    return false;
                }
            }
            continue
        } else if (key === '$count') {
            let op = get_op_func(condition[key]['op'])
            let value = condition[key]['value']
            let count = 0;
            for (let c of condition[key]['cond']) {
                if (match_condition(context, c)) {
                    count += 1;
                }
            }
            return op(count, value)
        }
        // 如果是对象，递归调用
        if (typeof condition[key] == "object" && condition[key] !== null) {
            if (!(key in context)) {
                return false;
            }
            if (!match_condition(context[key], condition[key])) {
                return false;
            }
        } else if (condition[key]) {
            if (condition[key]==='[无]'){
                //context中需要不包含这个key
                if(key in context){
                    return false;
                }
                continue;
            }
            if (condition[key] === context[key]) {
                continue
            }
            return false;
        } else if (!condition[key] && !context[key]) {
            continue
        } else {
            return false;
        }
    }
    return true;
}

function get_path_value(profile, path) {
    // 通过path获取profile中的值
    // path是一个字符串，用.分割，例如path_acne.tzone_acne
    let key_name = false;
    if (path.startsWith('[key_name]')){
        key_name = true
        path = path.slice(10)
    }
    for (let key of path.split('.')) {
        profile = profile[key]
    }
    if (key_name){
        for (let key in profile){
            if (profile[key]){
                return key
            }
        }
    }
    return profile
}

function update_context(context, obj) {
    // 更新上下文
    // context是上下文对象，obj是一个对象，用于更新上下文
    for (let key in obj) {
        // 如果 obj[key] 是一个对象，递归调用
        if (typeof obj[key] == "object") {
            if (!context[key]) {
                context[key] = {}
            }
            update_context(context[key], obj[key])
            continue
        } else {
            context[key] = obj[key]
        }
    }
}

function get_op_func(op_type) {
    // 获取操作函数
    // op_type是一个字符串，表示操作类型
    // 返回一个函数
    return {
        'eq': (a, b) => a == b,
        'neq': (a, b) => a != b,
        'gt': (a, b) => (a > b)&&(a!==null)&&(b!==null),
        'gte': (a, b) => (a >= b)&&(a!==null)&&(b!==null),
        'lt': (a, b) => (a < b)&&(a!==null)&&(b!==null),
        'lte': (a, b) => (a <= b)&&(a!==null)&&(b!==null),
        'and': (a, b) => {
            for (let c of b) {
                if (!get_path_value(a, c)) {
                    return false
                }
            }
            return true;
        },
        'or': (a, b) => {
            for (let c of b) {
                if (get_path_value(a, c)) {
                    return true
                }
            }
            return false;
        }
    } [op_type]
}

function init_context(context, profile_condition) {
    /*
        profile_condition 单个条件的结构
        {
        "name": "T区",
        "path": "path_acne.tzone_acne",
        "op_type": "lte",
        "only_one":true,
        "value": [{
                "name": "无",
                "value": 0
            },
            {
                "name": "轻",
                "value": 400
            },
            {
                "name": "中",
                "value": 800
            }, {
                "name": "重",
                "value": 1000
            }
        ]
    }
    计算过程：
    1. 从profile中获取{path}的值
    2. 选择 op_type 对应的操作函数
    3. 如果only_one为true，那么只要有一个条件满足，就不再继续计算，如果false，那么继续计算
    4. 操作函数(取出的值,value),如果返回true，那么就把这个条件的name加入到context.condition中
    */
    for (let condition of profile_condition) {
        let value = get_path_value(context, condition.path)
        let op = get_op_func(condition.op_type)
        if (!context.condition[condition.name]) {
            context.condition[condition.name] = {}
        }
        let have_one = false;
        let check_name = ''
        for (let v of condition.value) {
            let lop = op;
            if (v.op_type) {
                lop = get_op_func(v.op_type)
            }
            if (!(have_one && condition.only_one) && lop(value, v.value)) {
                context.condition[condition.name][v.name] = true
                check_name = v.name
                have_one = true; // 已经找到一个
            } else {
                if (check_name != v.name) {
                    context.condition[condition.name][v.name] = false
                }
            }
        }
    }
}
function make_text_response(context, text) {
    let reg = /{{(.*?)}}/g
    let match = reg.exec(text)
    while (match) {
        let key = match[1]
        // 如果key 里面包含.，那么就要从根获取路径数据
        let value = ''
        if (key.indexOf('.') != -1) {
            value = get_path_value(context, key)
        } else {
            value = get_path_value(context.user, key)
        }
        text = text.replace(match[0], value)
        match = reg.exec(text)
    }
    return text
}
function make_obj_response(context,obj){
    // 先判断obj是不是对象
    if (!obj){
        return obj
    }
    if (typeof(obj) == 'string'){
        return make_text_response(context,obj)
    }
    if (Array.isArray(obj)){
        for (let i=0;i<obj.length;i++){
            obj[i] = make_obj_response(context,obj[i])
        }
        return obj
    }
    for (let key in obj){
        if (typeof(obj[key]) == 'string'){
            obj[key] = make_text_response(context,obj[key])
        }else if (Array.isArray(obj[key])){
            for (let i=0;i<obj[key].length;i++){
                obj[key][i] = make_obj_response(context,obj[key][i])
            }
        }else if (typeof(obj[key]) == 'object'){
            obj[key] = make_obj_response(context,obj[key])
        }
    }
    return obj
}
function make_response(context, response) {
    // 将raw数据特殊处理完成后返回
    // 如果response是一个对象，则直接返回
    // 把context.user里面的内容填充进去
    // response 里面匹配{{}}里面的key，然后替换成context.user里面的值
    if(!response){
        return response
    }
    response = copy(response)
    if (typeof(response) == 'string'){
        response = {
            type:'text',
            value:response
        }
    }
    if (response.type === 'archives'){
        context.有档案 = true // 如果进入到档案，那么就设置有档案，为了便于二次进入时候判断
        let local_context = copy(context)
        local_context.咨询过程中 = true
        response.value = local_context
    }
    else{
        response = make_obj_response(context,response)
    }
    return response
}

function copy(obj) {
    // 深拷贝
    if (obj){
        return JSON.parse(JSON.stringify(obj))
    }
    return obj
}
function make_func(profile, screenplays, profile_condition,make_his_log=false) {
    let context = {
        init_user:copy(profile),
        user: copy(profile),
        current_screenplay: null, // 当前剧本，从0开始（0不是剧本）
        now_series: "1",
        choice: {
            // 记录用户选择的历史
        },
        play: {
            // 记录用户查看过的历史
        },
        condition: {
            // 计算过的用户的条件
        }
    }
    init_context(context, profile_condition)
    // 深度拷贝context 并存入history
    let history = []
    function set_profile_info(new_context){
        update_context(context.user,new_context)
    }
    function get_next() {
        if (make_his_log){
            history.push(JSON.parse(JSON.stringify(context)))
        }
        if (context.now_series == 'end') {
            return
        }
        // 如果前一个剧本有序列转移，则在这个时候生效
        if (context.current_screenplay && context.current_screenplay.to) {
            context.now_series = context.current_screenplay.to
        }
        // 如果一个剧本存在update_profile，那么就更新用户的profile
        if (context.current_screenplay && context.current_screenplay.update_profile) {
            update_context(context, context.current_screenplay.update_profile)
        }
        for (let screenplay of screenplays) {
            if (context.now_series != screenplay.series) {
                continue // 需要系列一致才考虑条件是否满足
            }
            if (match_condition(context, screenplay.condition)) {
                // 如果剧本已经播放过，那么就不再播放
                if (context.play[screenplay.key]) {
                    continue
                }
                context.current_screenplay = screenplay
                context.play[context.current_screenplay.key] = true
                // 如果screenplay.question是列表，那么需要通过条件匹配之后选择输出的问题
                let question = ""
                if (screenplay.question){
                    if (Array.isArray(screenplay.question)) {
                        question = ""
                        for (let q of screenplay.question) {
                            if (match_condition(context, q.condition)) {
                                question = make_response(context, q.text)
                                break
                            }
                        }
                    } else {
                        question = make_response(context, screenplay.question)
                    }
                }
                let options = []
                if (screenplay.options) {
                    // 删选掉不满足条件的option
                    for (let option of screenplay.options) {
                        if (!option.condition || match_condition(context, option.condition)) {
                            options.push({
                                "key":option.key,
                                "text":make_response(context, option.text),
                                "group":option.group
                            })
                        }
                    }
                }

                return copy({
                    'key': screenplay.key, // 剧本的key
                    "question": question||null, // 问题，是一个字符串或者对象，如果是对象，那么就要根据对象的type展示不同的界面
                    "options": options, // 选项列表 [{"key":A,"text":"选项A","show":true},{"key":B,"text":"选项B","show":false}]
                    "type": screenplay.type, // "multi":多选，"Critical Section":互斥分组选择
                    "pre_ans": make_response(context, screenplay.pre_ans), // 用户回答的前置回答
                    "group_type": screenplay.group_type, // 描述组内的选择类型
                })

            }
        }
        if (context.now_series != 'end') {
            console.log('没找到剧本意外结束')
            console.log(context)
            return {'type':"error",'value':"没找到剧本意外结束"}
        }
    }

    function set_select(select) {
        // select 是一个数组，每个条目是选项的key
        // 记录用户选择的历史
        if (make_his_log){
            history.push(JSON.parse(JSON.stringify(context)))
        }
        for (let key of select) {
            context.choice[key] = true
        }
        // 创建反馈
        let response = []
        if (context.current_screenplay.options) {
            for (let option of context.current_screenplay.options) {
                if (select.includes(option.key)) {
                    if (option.response) {
                        // 如果option.response 是列表，那么就填充写入，否则就直接录入
                        let local_resp = option.response
                        if (!Array.isArray(local_resp)) {
                            local_resp = [local_resp]
                        }
                        for (let l_resp of local_resp) {
                            response.push(make_response(context, l_resp))
                        }
                    }
                    if (option.to) {
                        context.now_series = option.to
                    }
                    if (option.update_profile) { // 更新context
                        update_context(context, option.update_profile)
                    }
                }
            }
        }
        // 如果有end，追加end
        if (context.current_screenplay.end) {
            // 如果是数组，那么需要通过条件匹配之后选择输出的结束语
            let end = context.current_screenplay.end
            if (Array.isArray(end)) {
                for (let e of end) {
                    if (match_condition(context, e.condition)) {
                        end = e.text
                        response.push(make_response(context, end))
                        break
                    }
                }
            } else {
                response.push(make_response(context, end))
            }
        }
        return response
    }

    function get_context() {
        return context
    }

    function go_back() {
        if (history.length >= 1) {
            for (let key in history[history.length - 1]) {
                context[key] = history[history.length - 1][key]
            }
            for (let key in context){
                if (key in history[history.length - 1]){
                    continue
                }
                delete context[key]
            }
            history.pop()
        }
    }
    return {
        get_next,
        set_select,
        get_context,
        set_profile_info,
        go_back
    }
}
function get_archives_value(item,context){
    let res = []
    if (match_condition(context,item.condition)){
        if (item.sub_list){
            for (let sub_item of item.sub_list){
                res = res.concat(get_archives_value(sub_item,context))
            }
            return res
        }else{
            if (typeof(item.value) == 'string'){
                for (let line of item.value.split('\n')){
                    if (line.length>0){
                        res.push(make_text_response(context,line))
                    }
                }
                return res
            }else{
                return [make_obj_response(context,item.value)]
            }
        }
    }
    return res
}
// 导出make_core_func
    function make_archives(context,archives_json){
        // 通过上下文已经档案json生成档案信息并返回
        // if (!context.有档案){
        //     return
        // }
        let res = {}
        res = copy(archives_json['default'])
        for (let key in res){
            if (!Array.isArray(res[key])){
                res[key] = this.make_archives(context,{'default':res[key],'content':archives_json['content'][key]})
            }else{
                for (let item of archives_json['content'][key]){
                    for (let value of get_archives_value(item,context)){
                        res[key].push(value)
                    }
                }
            }
        }
        return res;
    }
    function make_core_func_with_tools(profile, screenplays, profile_condition){
        return make_func(profile, screenplays, profile_condition,true)
    }
    function make_core_func(profile, screenplays, profile_condition) {
        // 封装 make_core_func_with_tools,只返回get_next和set_select
        let func = make_func(profile, screenplays, profile_condition)
        return {
            get_next: func.get_next,
            set_select: func.set_select,
            get_context: func.get_context,
            set_profile_info:func.set_profile_info,
        }
    }
